#include <iostream>
#include "Bit_stream.h"
#include "xma_parse.h"

using namespace std;


unsigned int Parse_XMA::parse_block(istream& is, ostream& os, long offset, long block_size, int channels, bool verbose) {
    long last_offset = offset + block_size;
    unsigned long sample_count = 0;

    for (int packet_number = 0; offset < last_offset; packet_number++) {
        packet_header ph;

        {
            Bit_stream packet_header_stream(is);

            is.seekg(offset);

            packet_header_stream >> ph;
        }

        if (verbose) {
            cout << "Packet #" << packet_number << " (offset " << hex << offset << dec << ")" << endl;
            cout << "Frame Count     " << ph.frame_count << endl;
            cout << "Skip Bits       " << ph.skip_bits << endl;
            cout << "Metadata        " << ph.metadata << endl;
            cout << "Packet Skip     " << ph.packet_skip << endl;
        }

        Bit_stream frame_stream(is,
                // consecutive
                (packet_size_bytes - packet_header_size_bytes) * 8,
                // skip
                (packet_header_size_bytes + ph.packet_skip * packet_size_bytes) * 8
                );

        // skip initial bits (overflow from a previous packet)
        for (unsigned int i = 0; i < ph.skip_bits; i++) frame_stream.get_bit();

        bool packet_end_seen = false;

        for (unsigned int frame_number = 0; frame_number < ph.frame_count; frame_number++) {
            Bit_uint<frame_header_size_bits> frame_bits;
            frame_stream >> frame_bits;
            unsigned int bits_left = frame_bits - frame_header_size_bits;

            if (verbose)
            {
                cout << "   Frame #" << frame_number << endl;
                cout << "   Size " << frame_bits << endl;
            }

            // sync?
            {
                Bit_uint<frame_sync_size_bits> sync;
                frame_stream >> sync;
                if (sync!= 0x7f00) throw Bad_frame_sync(sync);
                bits_left -= frame_sync_size_bits;
            }

            if (channels > 1) {
                frame_stream.get_bit();
                bits_left --;
            }

            // skip
            if (frame_stream.get_bit()) {
                Bit_uint<frame_skip_size_bits> skip_start, skip_end;
                // skip at start
                if (frame_stream.get_bit()) {
                    frame_stream >> skip_start;
                    if (verbose) {
                        cout << "Skip " << skip_start << " samples at start" << endl;
                    }
                    bits_left -= frame_skip_size_bits;
                    sample_count -= skip_start;
                }
                bits_left --;
                // skip at end
                if (frame_stream.get_bit()) {
                    frame_stream >> skip_end;
                    if (verbose) {
                        cout << "Skip " << skip_end << " samples at end" << endl;
                    }
                    bits_left -= frame_skip_size_bits;
                    sample_count -= skip_end;
                }
                bits_left --;
            }
            bits_left --;

            if (verbose) {
                cout << hex;
            }
            for (; bits_left >= 4 + frame_trailer_size_bits; bits_left -= 4) {
                Bit_uint<4> nybble;
                frame_stream >> nybble;
                if (verbose) {
                    cout << nybble;
                }
            }
            if (verbose) {
                cout << " ";
            }
            for (; bits_left > frame_trailer_size_bits; bits_left--) {
                bool bit = frame_stream.get_bit();
                if (verbose) {
                    cout << (bit ? '1' : '0');
                }
            }
            if (verbose) {
                cout << dec << endl;
            }

            // trailer
            {
                if (!frame_stream.get_bit())
                {
                    if (frame_number != ph.frame_count-1) throw Early_packet_end();
                    packet_end_seen = true;
                }

                sample_count += samples_per_frame;

                bits_left -= frame_trailer_size_bits;
            }
        }

        // Occasionally the previous packet had frames so long that we have to skip all
        // of this packet!
        if (ph.frame_count == 0) {
            if (ph.skip_bits != 0x7fff) {
                throw Zero_frames_not_skipped();
            }
            packet_end_seen = true;
        }

        if (!packet_end_seen) throw Missing_packet_end();

        if (verbose) {
            cout << endl;
        }

        // We've successfully examined this packet, dump it out if we have an output stream
        if (os) {
            char buf[packet_size_bytes];
            is.seekg(offset);

            is.read(buf, packet_size_bytes);
            buf[3] = 0; // zero packet skip, since we're packing consecutively
            os.write(buf, packet_size_bytes);
        }

        // advance to next packet
        offset += (ph.packet_skip + 1) * packet_size_bytes;
    }

    return sample_count;
}
